# pyright: reportUnusedFunction=false
from typing import cast

import jinja2
import markdown
from jinja2.utils import markupsafe  # type: ignore
from klein import Klein
from klein.resource import KleinResource
from twisted.logger import Logger
from twisted.python.filepath import FilePath
from twisted.web import resource, static
from twisted.web.server import Request

from .AdlerManagerTokenResource import AdlerManagerTokenResource
from .Config import Config
from .SitesManager import SiteManager, SitesManager

log = Logger()


def get_jinja_env(supportDir: str) -> jinja2.Environment:
    """
    Return a L{jinja2.Environment} with templates loaded from:
      - Package
      - Support dir

    @param supportDir: Full path to supportDir.
      See L{authapiv02.DefaultConfig.Config}
    @type supportDir: L{str}
    """
    md = markdown.Markdown(
        extensions=[
            "markdown.extensions.toc",
            "markdown.extensions.tables",
        ]
    )
    templates = jinja2.Environment(
        extensions=["jinja2.ext.do", "jinja2.ext.loopcontrols"],
        loader=jinja2.ChoiceLoader(
            [
                jinja2.FileSystemLoader(supportDir),
                jinja2.PackageLoader("adlermanager", "templates"),
            ]
        ),
        autoescape=True,
    )

    def md_filter(txt: str) -> markupsafe.Markup:
        return markupsafe.Markup(md.convert(txt))

    templates.filters["markdown"] = md_filter  # type: ignore
    return templates


def web_root(sites_manager: "SitesManager") -> KleinResource:
    app = Klein()

    @app.route("/")  # type: ignore
    def index(request: Request):
        try:
            host = cast(str, request.getRequestHostname().decode("utf-8"))
        except Exception:
            return resource.ErrorPage(
                400, "Bad cat", '<a href="http://http.cat/400">http://http.cat/400</a>'
            )
        if host not in sites_manager.site_managers:
            return resource.ErrorPage(
                404, "Gone cat", '<a href="http://http.cat/404">http://http.cat/404</a>'
            )
        site: SiteManager
        try:
            site = sites_manager.site_managers[host]
        except Exception:
            log.failure("sad cat")
            return resource.ErrorPage(
                500, "Sad cat", '<a href="http://http.cat/500">http://http.cat/500</a>'
            )

        site_path = cast(  # type: ignore
            str, FilePath(Config.data_dir).child("sites").child(host).path
        )
        templates = get_jinja_env(site_path)
        template = templates.get_template("template.j2")

        return template.render(site=site)

    @app.route("/api/v1/alerts", methods=["POST"])  # type: ignore
    def alert_handler(request: Request):
        return AdlerManagerTokenResource(sites_manager)

    @app.route("/static", branch=True)  # type: ignore
    def static_files(request: Request):
        return static.File(Config.web_static_dir)

    return app.resource()



from datetime import datetime, timezone

import attr
from twisted.internet import defer
from twisted.python.failure import Failure
from twisted.python.filepath import FilePath

_blessed_date_format = "%Y-%m-%dT%H:%M:%S%z"


@attr.s
class TimestampFile(object):
    path: FilePath = attr.ib()

    def set(self, time: datetime) -> None:
        with self.path.open("w") as f:
            f.write(time.strftime(_blessed_date_format).encode("utf-8"))

    def now(self) -> None:
        self.set(current_time())

    def getStr(self) -> str:
        if not self.path.exists():
            return ""
        with self.path.open("r") as f:
            return f.read().decode("utf-8")


def current_time() -> datetime:
    return datetime.now(timezone.utc)


def current_timestamp() -> str:
    return current_time().strftime(_blessed_date_format)


def read_timestamp(s: str) -> datetime:
    # We drop nanoseconds as python does not support that
    return datetime.strptime(f"{s.split('.')[0]}+00:00", _blessed_date_format)


def ensure_dirs(path: FilePath) -> None:
    path.makedirs(ignoreExistingDirectory=True)


def default_errback(failure: Failure) -> None:
    failure.trap(defer.CancelledError)  # type: ignore
    pass


def noop_deferred() -> defer.Deferred[None]:
    d: defer.Deferred[None] = defer.Deferred()
    _ = d.addErrback(default_errback)
    return d



import os
from datetime import timedelta

import attr


@attr.s
class ConfigClass(object):
    """
    All settings can be set via Environment variables.
    If you use pipenv, you are encouraged to use the .env file.
    See pipenv's documentation for more information.
    """

    data_dir: str = attr.ib(default=os.getenv("DATA_DIR", "../data"))
    """
    @param data_dir: Environment: DATA_DIR.
           Directory to save persistent data in.
    @type  data_dir: C{unicode}
    """

    # Web
    web_endpoint: str = attr.ib(
        default=os.getenv("WEB_ENDPOINT", r"tcp6:interface=\:\::port=8080")
    )
    """
    @param web_endpoint: Environment: WEB_ENDPOINT.
           Where we should listen for HTTP connections.
           This defaults to TCP port 8080, but you can also use a UNIX socket,
           which you can proxy from your web server.

           See:
           https://docs.twisted.org/en/stable/core/howto/endpoints.html#servers
           https://klein.readthedocs.io/en/latest/examples/alternativerunning.html#example-ipv6-tls-unix-sockets-endpoints
    @type  web_endpoint: C{unicode} -- Endpoint string
    """

    web_static_dir: str = attr.ib(default=os.getenv("WEB_STATIC_DIR", "../static"))
    """
    @param web_static_dir: Environment: WEB_STATIC_DIR.
           Directory to server static files from.
    @type  web_static_dir: C{unicode}
    """

    # SSH
    ssh_enabled: bool = attr.ib(default=os.getenv("SSH_ENABLED", "YES") != "")
    """
    @param ssh_enabled: Environment: SSH_ENABLED.
           If this environment variable is anything other than empty,
           SSH_ENDPOINT will be used to listen for SSH connections.
    @type  ssh_enabled: C{unicode}
    """

    ssh_endpoint: str = attr.ib(
        default=os.getenv("SSH_ENDPOINT", r"tcp6:interface=\:\::port=2222")
    )
    """
    @param ssh_endpoint: Environment: SSH_ENDPOINT.
           Where we should listen for SSH connections if SSH_ENABLED is
           anything other than empty.

           See:
           https://docs.twisted.org/en/stable/core/howto/endpoints.html#servers
           https://klein.readthedocs.io/en/latest/examples/alternativerunning.html#example-ipv6-tls-unix-sockets-endpoints
    @type  ssh_endpoint: C{unicode} -- Endpoint string
    """

    ssh_key_size: int = attr.ib(default=int(os.getenv("SSH_KEY_SIZE", "4096")))
    """
    @param ssh_key_size: Environment: SSH_KEY_SIZE.
           Size for server's auto-generated SSH host key.
    @type  ssh_key_size: C{unicode}
    """

    ssh_keys_dir: str = attr.ib(default=os.getenv("SSH_KEYS_DIR", "../data/ssh"))
    """
    @param ssh_keys_dir: Environment: SSH_KEYS_DIR.
           Directory to save SSH keys in.
           This includes the server private key and users' public keys.
    @type  ssh_keys_dir: C{unicode}
    """

    # Alerts processing
    alert_resolve_minutes: timedelta = attr.ib(
        default=timedelta(minutes=int(os.getenv("ALERT_RESOLVE_MINUTES", "5")))
    )
    """
    @param alert_resolve_minutes: Environment: ALERT_RESOLVE_MINUTES.
           Prometheus keeps sending alerts periodically as long as it is aware
           of them, and when it determines they have been resolved, it too
           sends us that information.
           However, under certain circumstances, an alert may be raised and it
           can just stop being sent.
           This variable marks how long we should remember such alerts that
           stop arriving but are not explicitly marked as resolved.
           Default value: 5 (minutes), as AlertManager
    """

    group_incidents_minutes: timedelta = attr.ib(
        default=timedelta(minutes=int(os.getenv("GROUP_INCIDENTS_MINUTES", "60")))
    )
    """
    @param group_incidents_minutes: Environment: GROUP_INCIDENTS_MINUTES.
           Alerts incoming at roughly the same time are grouped into incidents.
           Use this to configure the timeout after which alerts will be
           considered separate incidents, in minutes.
           Example: If it is set for 30 minutes, and alerts come at 09:00,
           09:20, 09:40, and 11:00, then the first 3 will be the same incident and
           the last one will begin a new incident.
           Default value: 60 (i.e. 1 hour).
    """

    monitoring_down_minutes: timedelta = attr.ib(
        default=timedelta(minutes=int(os.getenv("MONITORING_DOWN_MINUTES", "2")))
    )
    """
    @param monitoring_down_minutes: Environment: MONITORING_DOWN_MINUTES.
           We expect prometheus to get in touch with AdlerManager periodically.
           In order to achieve this, a special rule should be created (and
           silenced in AlertManager!), which acts as a "Heartbeat" signal.
           If we do not do this, monitoring may be down and we'll never get
           see that.
           The rule should look like this:

           # groups:
           # - name: "AdlerManager at example.org"
           # interval: 30s
           # rules:
           # - alert: "AdlerManager: Everything is fine"
           #     # This expression should always be true.
           #     # A neat trick is using the job that gathers prometheus metrics
           #     expr: up{job="prometheus"} == 1
           #     labels:
           #         # This must match the sitename in AdlerManager
           #         adlermanager: "example.org"
           #         # This must be set
           #         heartbeat: True
           #         # This is optional, but it is nice to have
           #         severity: OK


           This variable determines how long we can go without having contact
           from Prometheus. When this time is exceeded, AdlerManager will
           fallback to a "Monitoring is Down" state.
           Default value: 2 (minutes)
    """


Config = ConfigClass()

if __name__ == "__main__":
    import inspect
    import sys

    target = "dotenv.example"
    if len(sys.argv) > 1:
        target = sys.argv[1]

    sc = inspect.getsource(ConfigClass).split("\n")[2:]

    with open(target, "w") as f:
        f.write(
            "\n".join(
                [
                    "#",
                    "# Automatically generated, manual changes will be lost",
                    "# Run: python -m adlermanager.Config dotenv.example",
                    "#",
                    "# If using Pipenv or similar, you probably want:",
                    "# To load the adlermanager module:",
                    'PYTHONPATH="./src"',
                    "# To instruct mypy to find the adlermanager module:",
                    'MYPYPATH="./src"',
                    "",
                    "",
                ]
            )
        )
        in_doc = False
        for line in sc:
            if line.strip() == '"""':
                in_doc = not in_doc
                continue
            if "@type" in line:
                continue

            if in_doc or line.strip().startswith("#"):
                if "@param" in line:
                    f.write(f"# {line.split(':', 1)[1].strip()}")
                else:
                    f.write(f"# {line[min(4, len(line)):].strip()}".strip())
                f.write("\n")
            if not in_doc and "getenv" in line:
                parts = line.split('"')
                f.write("\n")
                f.write(f'#{parts[1]}="{parts[3]}"')
                f.write("\n#\n")



from io import BytesIO
from typing import (
    Any,
    Callable,
    Coroutine,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
    cast,
)

from zope.interface import implementer

from twisted.application import strports
from twisted.application.internet import StreamServerEndpointService
from twisted.conch import avatar, interfaces as conchinterfaces, manhole_ssh, recvline
from twisted.conch.checkers import (
    IAuthorizedKeysDB,
    SSHPublicKeyChecker,
    readAuthorizedKeyFile,
)
from twisted.conch.insults import insults
from twisted.conch.manhole_tap import chainedProtocolFactory
from twisted.conch.ssh import keys, session
from twisted.cred import portal
from twisted.internet import defer
from twisted.internet.error import ProcessTerminated
from twisted.python import failure, filepath


class SSHSimpleProtocol(recvline.HistoricRecvLine):
    terminal: insults.ServerProtocol
    keyHandlers: Dict[bytes, Callable[[], None]]
    mode: str
    lineBuffer: List[bytes]
    _unicodeCharBuffer: List[bytes]
    _unicodeCharLength: int

    _command_lineReceived: Optional[Callable[[bytes], None]] = None
    _command_inputEnded: Optional[Callable[[bytes], None]] = None

    def __init__(self, user: "SSHSimpleAvatar") -> None:
        recvline.HistoricRecvLine.__init__(self)
        self.user = user
        self.ps = (b">>> ", b"... ")
        self._unicodeCharBuffer = []
        self._unicodeCharLength = 1

    def terminal_write(self, msg: Union[bytes, str]) -> None:
        """
        Write data to the remote terminal.

        We use this method to work around awkward typing in t.c.insults.
        """
        return self.terminal.write(msg)  # type: ignore

    def initializeScreen(self) -> None:
        """
        Initialise terminal in InsertMode, without clearing the screen.

        This overrides the definition in twisted.conch.recvline.RecvLine:
        https://github.com/twisted/twisted/blob/7697871b4d89c78c8764c6be42372fc68299714e/src/twisted/conch/recvline.py#L385
        """
        # This seems to produce minor issues on the client's terminal
        # self.setInsertMode()
        # But conch does assume .mode is set early
        self.mode = "insert"
        if self.interactive:
            self.terminal_write(self.motd())
            self.terminal.nextLine()
            self.showPrompt()

    def keystrokeReceived(self, keyID: bytes, modifier: Any) -> None:
        """
        Handle a received keystroke.

        The default implementation of this was keeping us from using anything
        not listed in string.printable, which is a rather limited subset of
        symbols.
        """
        # For those of us that forget periodically how unicode works:
        # https://docs.python.org/3/library/codecs.html#encodings-and-unicode
        # The table with the encoding is particularly useful :-)
        unicode_mark = 1 == (keyID[0] >> 7) & 1  # 0x1XXXYYYY
        m = self.keyHandlers.get(keyID)
        if m is not None:
            m()
        elif (
            not unicode_mark  # This is not a unicode byte
            and not self._unicodeCharBuffer  # Check we had nothing in buffer
            and keyID in self._printableChars  # Printable ascii
        ):
            self.characterReceived(keyID, False)  # type: ignore
        elif unicode_mark:  # Handle unicode
            leading_ones = 1  # 0x10XXYYYY .. 0x11110YYY
            for i in (6, 5, 4, 3):
                if 1 == keyID[0] >> i & 1:
                    leading_ones += 1
                else:
                    break
            if (  # Catch misbehaved input
                leading_ones > 4  # This is too many ones
                or (
                    leading_ones < 2 and len(self._unicodeCharBuffer) < 1
                )  # This is a wrong starting byte
                or (
                    leading_ones > 1 and len(self._unicodeCharBuffer) > 0
                )  # This  ends a previous multi-byte char too soon
                # TODO: should we keep this starting byte in the buffer?
            ):
                self._unicodeCharBuffer.append(keyID)
                self._log.warn(
                    "Received unhandled unicode sequence: {bs!r}",
                    bs=b"".join(self._unicodeCharBuffer),
                )
                self._unicodeCharBuffer.clear()
            else:
                if leading_ones > 1:  # First byte of char
                    self._unicodeCharBuffer.clear()
                    # Save how many bytes we need
                    self._unicodeCharLength = leading_ones

                self._unicodeCharBuffer.append(keyID)
                if len(self._unicodeCharBuffer) == self._unicodeCharLength:
                    # Got enough bytes, send the char
                    self.characterReceived(  # type: ignore
                        b"".join(self._unicodeCharBuffer), False
                    )
                    self._unicodeCharBuffer.clear()
        else:
            self._log.warn("Received unhandled keyID: {keyID!r}", keyID=keyID)
            if self._unicodeCharBuffer:
                self._log.warn(
                    "    after unicode sequence: {bs!r}",
                    bs=b"".join(self._unicodeCharBuffer),
                )
            self._unicodeCharBuffer.clear()

        if not unicode_mark:
            self._unicodeCharBuffer.clear()

    def connectionMade(self) -> None:
        recvline.HistoricRecvLine.connectionMade(self)
        if not self.interactive:
            return
        # CTRL_D
        self.keyHandlers[b"\x04"] = self.handle_EOF
        # CTRL_BACKSLASH
        self.keyHandlers[b"\x1c"] = self.handle_QUIT

    def handle_EOF(self) -> None:
        if self._command_inputEnded is not None:
            self._command_inputEnded(b"".join(self.lineBuffer))
        if self.lineBuffer:  # type: ignore
            self.terminal_write(b"\a")
        else:
            self.exitWithCode(0)

    def handle_QUIT(self) -> None:
        self.terminal.loseConnection()

    @property
    def interactive(self) -> bool:
        """
        Whether or not we are running an interactive shell.

        This is currently done by determining if a pty has been requested.
        """
        return self.user.ptyAllocated

    @property
    def prompt(self) -> bytes:
        return self.ps[self.pn]

    def showPrompt(self) -> None:
        if self.interactive:
            self.terminal_write(self.prompt)

    def _getCommand(
        self, cmd: bytes
    ) -> Optional[Callable[..., Union[None, Coroutine[Any, Any, None]]]]:
        """
        Get the method that would be run by 'cmd' if appliable.

        The convention is that a cmd command translates to a do_cmd method.
        """
        try:
            cmd_str = cmd.decode("utf-8")
            return getattr(self, "do_" + cmd_str, None)
        except Exception:
            return None

    async def runCommand(self, cmd: bytes, *args: bytes) -> None:
        """
        Run the requested command or print 'No such command'.
        """
        func = self._getCommand(cmd)
        if func:
            try:
                res = func(*args)
                if res is not None:
                    await defer.ensureDeferred(res)
                if not self.interactive:
                    self.exitWithCode(0)
            except Exception as ex:
                self.terminal_write("Error: {}".format(ex))
                self.terminal.nextLine()
                if not self.interactive:
                    self.exitWithCode(2)
        else:
            self.terminal_write(b"No such command: " + cmd)
            self.terminal.nextLine()
            if not self.interactive:
                self.exitWithCode(1)

    async def runLine(self, line: bytes) -> None:
        line_parts = line.strip().split()
        if line_parts:
            cmd, *args = line_parts
            return await self.runCommand(cmd, *args)

    async def _lineReceived(self, line: bytes) -> None:
        await self.runLine(line)
        self.showPrompt()

    def lineReceived(self, line: bytes) -> None:
        if self._command_lineReceived is None:
            _ = defer.ensureDeferred(self._lineReceived(line))
        else:
            self._command_lineReceived(line)

    def get_user_input(self, eom: bytes = b"") -> defer.Deferred[bytes]:
        d = defer.Deferred[bytes]()
        data = BytesIO()

        def _ui_lineReceived(line: bytes) -> None:
            if eom and line != eom:
                data.write(line)
                data.write(b"\n")
            else:
                _ui_inputEnded(line)

        def _ui_inputEnded(line: bytes) -> None:
            if line and line != eom:
                data.write(line)
                data.write(b"\n")
            self._command_lineReceived, self._command_inputEnded = None, None
            d.callback(data.getvalue())

        self._command_lineReceived, self._command_inputEnded = (
            _ui_lineReceived,
            _ui_inputEnded,
        )
        return d

    def exitWithCode(self, code: int) -> None:
        cast(session.SSHSessionProcessProtocol, self.terminal.transport).processEnded(
            failure.Failure(ProcessTerminated(code, None, None))
        )

    def do_help(self, cmd: bytes = b"") -> None:
        """
        Get help on a command. Usage: help [command]
        """
        if cmd:
            func = self._getCommand(cmd)
            if func:
                self.terminal_write(
                    getattr(func, "__doc__", b"No help for '" + cmd + b"'")
                )
            else:
                self.terminal_write(b"No such command: " + cmd)
        else:
            self.terminal_write("Available commands:")
            self.terminal.nextLine()
            self.terminal.nextLine()
            for c in (attr[3:] for attr in dir(self) if attr.startswith("do_")):
                self.terminal_write(c)
                self.terminal.nextLine()
        self.terminal.nextLine()

    def do_whoami(self) -> None:
        """
        Prints your username. Usage: whoami
        """
        self.terminal_write(self.user.username)
        self.terminal.nextLine()

    def do_clear(self) -> None:
        """
        Clears the screen. Usage: clear
        """
        self.terminal.reset()

    def do_exit(self) -> None:
        """
        Exit session. Usage: exit
        """
        self.exitWithCode(0)

    def motd(self) -> Union[str, bytes]:
        return ""


@implementer(conchinterfaces.ISession)
class SSHSimpleAvatar(avatar.ConchUser):
    serverProtocol: Optional[insults.ServerProtocol] = None
    ptyAllocated: bool = False
    """
    Whether or not a pty was ever requested.
    """

    def __init__(self, username: bytes, proto: SSHSimpleProtocol):
        avatar.ConchUser.__init__(self)

        self.username = username
        self.proto = proto
        self.channelLookup.update({b"session": session.SSHSession})  # type: ignore

    def openShell(self, protocol: session.SSHSessionProcessProtocol) -> None:
        self.serverProtocol = insults.ServerProtocol(self.proto, self)
        self.serverProtocol.makeConnection(protocol)  # type: ignore
        protocol.makeConnection(  # type: ignore
            session.wrapProtocol(self.serverProtocol)  # type: ignore
        )

    def getPty(self, terminal, windowSize, attrs) -> None:  # type: ignore
        # Save the pty request for use from the protocol
        self.ptyAllocated = True
        return None

    def execCommand(
        self, protocol: session.SSHSessionProcessProtocol, line: bytes
    ) -> None:
        self.serverProtocol = insults.ServerProtocol(self.proto, self)
        self.serverProtocol.makeConnection(protocol)  # type: ignore
        protocol.makeConnection(  # type: ignore
            session.wrapProtocol(self.serverProtocol)  # type: ignore
        )
        _ = defer.ensureDeferred(
            cast(
                SSHSimpleProtocol, self.serverProtocol.terminalProtocol  # type: ignore
            ).runLine(line)
        )

    def windowChanged(self, dimensions: Tuple[int, int, int, int]):  # type: ignore
        """This gets triggered when user changes terminal dimensions.

        Args:
            dimensions (Tuple[int, int, int, int]): Probably height, width and "0, 0"
        """
        pass

    def eofReceived(self) -> None:
        """
        Called when the other side has indicated no more data will be sent.
        """
        if self.serverProtocol is not None:
            ssh_sp = cast(
                SSHSimpleProtocol, self.serverProtocol.terminalProtocol  # type: ignore
            )
            ssh_sp.handle_EOF()

    def closed(self) -> None:
        """
        Called when the session is closed.
        """
        pass


@implementer(portal.IRealm)
class SSHSimpleRealm:
    """
    SSH simple realm that uses a given protocol for any valid users.

    You may want to customise the way that protocol works, realm and avatar
    are not necessarily of interest here.

    @ivar proto: The passed protocol class that will be used for avatar.
    """

    def __init__(self, proto: SSHSimpleProtocol):
        """
        Initialise a new L{SSHSimpleRealm} with proto as a protocol class.

        @param proto: a protocol class that will be used for any avatars.
        """
        self.proto = proto

    def requestAvatar(
        self,
        avatarId: Union[bytes, Tuple[()]],
        mind: object,
        *interfaces: portal._InterfaceItself,  # type: ignore
    ) -> Union[
        defer.Deferred[portal._requestResult], portal._requestResult  # type: ignore
    ]:
        """
        Return a L{SSHSimpleAvatar} that uses ``self.proto`` as protocol.

        @see: L{portal.IRealm}
        """
        if conchinterfaces.IConchUser in interfaces:
            avatar = SSHSimpleAvatar(cast(bytes, avatarId), self.proto)
            return interfaces[0], avatar, lambda: None
        else:
            raise Exception("No supported interfaces found.")


@implementer(IAuthorizedKeysDB)
class SSHKeyDirectory(object):
    """
    Provides SSH public keys based on a simple directory structure.

    For a user ``USER`` following files are returned if they exist:
      - ``$USER/*.key``
      - ``$USER.key``
    These paths are relative to L{SSHKeyDirectory.baseDir}

    @ivar baseDir: the base directory for key lookup.
    """

    def __init__(
        self,
        baseDir: filepath.FilePath[str],
        parseKey: Any = keys.Key.fromString,  # type: ignore
    ) -> None:
        """
        Initialises a new L{SSHKeyDirectory}.

        @param base_dir: the base directory for key lookup.
        @param parseKey: L{callable}
        """
        self.baseDir = baseDir
        self.parseKey = parseKey

    def getAuthorizedKeys(self, username: bytes) -> Iterator[keys.Key]:
        keyFile = self.baseDir.child(username + b".key")
        keyDir = self.baseDir.child(username)

        if keyFile.isfile():
            for key in readAuthorizedKeyFile(keyFile.open(), self.parseKey):
                yield key

        if keyDir.isdir():
            for f in keyDir.globChildren("*.key"):
                for key in readAuthorizedKeyFile(f.open(), self.parseKey):
                    yield key


def conch_helper(
    endpoint: str,
    proto: Optional[SSHSimpleProtocol] = None,
    namespace: Dict[str, str] = dict(),
    keyDir: Optional[str] = None,
    keySize: int = 4096,
) -> StreamServerEndpointService:
    """
    Return a L{SSHKeyDirectory} based SSH service with the given parameters.

    Authorized keys are read as per L{SSHKeyDirectory} with ``baseDir`` being
    ``keyDir/users``.

    @param endpoint: endpoint for the SSH service
    @param namespace: the manhole namespace
    @param keyDir: directory that holds server/server.key file and
        users directory, which is used as ``baseDir`` in L{SSHKeyDirectory}
    @see: L{SSHKeyDirectory}
    """
    if keyDir is None:
        from twisted.python._appdirs import getDataDirectory

        keyDir = getDataDirectory()

    ssh_keys_dir = filepath.FilePath(keyDir)
    ssh_keys_dir.child("server").makedirs(True)
    ssh_keys_dir.child("users").makedirs(True)

    checker = SSHPublicKeyChecker(
        SSHKeyDirectory(ssh_keys_dir.child("users"))  # type: ignore
    )

    if proto is None:
        sshRealm = manhole_ssh.TerminalRealm()
        sshRealm.chainedProtocolFactory = chainedProtocolFactory(  # type: ignore
            namespace
        )
    else:
        sshRealm = SSHSimpleRealm(proto)  # type: ignore
    sshPortal = portal.Portal(sshRealm, [checker])  # type: ignore

    sshKeyPath = ssh_keys_dir.child("server").child("server.key")
    sshKey = keys._getPersistentRSAKey(sshKeyPath, keySize)  # type: ignore

    sshFactory = manhole_ssh.ConchFactory(sshPortal)
    sshFactory.publicKeys[b"ssh-rsa"] = sshKey  # type: ignore
    sshFactory.privateKeys[b"ssh-rsa"] = sshKey  # type: ignore

    sshService = strports.service(endpoint, sshFactory)  # type: ignore
    return sshService



from datetime import datetime
from enum import IntEnum
from typing import Any, Dict, Optional, Union, cast

import attr
import yaml

from .utils import current_time, read_timestamp


class Severity(IntEnum):
    OK = 0
    INFO = 1
    WARNING = 2
    ERROR = 3

    @classmethod
    def from_string(cls, s: str) -> "Severity":
        # TODO: Do something sensitive with other priorities
        #       At least document them somewhere :-D
        labels = {
            "ok": cls.OK,
            "info": cls.INFO,
            "warning": cls.WARNING,
            "error": cls.ERROR,
            "critical": cls.ERROR,
        }
        return labels[s.lower()]

    @classmethod
    def from_alert(cls, alert: "Alert") -> "Severity":
        if alert.endsAt and alert.endsAt <= current_time():
            return Severity.OK
        return Severity.from_string(alert.labels.get("severity", "OK"))

    @property
    def css(self) -> str:
        classes = {self.OK: "success", self.WARNING: "warning", self.ERROR: "danger"}
        return classes[self]

    def __str__(self) -> str:
        return self.css


@attr.s
class Alert:
    labels: Dict[str, str] = attr.ib(factory=dict)
    annotations: Dict[str, str] = attr.ib(factory=dict)
    endsAt: Optional[datetime] = attr.ib(default=None)
    startsAt: Optional[datetime] = attr.ib(default=None)
    status: Severity = attr.ib(default=Severity.OK)

    @classmethod
    def import_alert(cls, d: Dict[str, Any]) -> "Alert":
        # Convert date data types
        alert = Alert(
            labels=d.get("labels", dict()),
            annotations=d.get("annotations", dict()),
        )
        for att in ["startsAt", "endsAt"]:
            if att in d:
                try:
                    setattr(alert, att, read_timestamp(cast(str, d[att])))
                except Exception:
                    setattr(alert, att, None)
        # Convert severity (needs date data)
        alert.status = Severity.from_alert(alert)
        return alert


@attr.s
class SiteConfig(object):
    message: str = attr.ib(default="")
    force_state: bool = attr.ib(default=False)

    @property
    def title(self) -> str:
        parts = self.message.split("\n\n")
        if len(parts) < 2:
            return ""
        return parts[0]

    @property
    def body(self) -> str:
        parts = self.message.split("\n\n")
        if len(parts) < 2:
            return self.message
        return parts[1]

    @property
    def state_is_forced(self) -> bool:
        if self.message is None or self.message == "":
            return False
        return self.force_state

    def to_YAML(self) -> str:
        return yaml.safe_dump(attr.asdict(self), allow_unicode=True)

    @classmethod
    def from_YAML(cls, yaml_string: Union[bytes, str]) -> "SiteConfig":
        """
        Create a SiteConfig object from a YAML string / bytes.

        Args:
            yaml_string (str): The YAML string representing our SiteConfig.

        Returns:
            SiteConfig: The SiteConfig represented by yaml_string.
        """
        obj = yaml.safe_load(yaml_string)
        return SiteConfig(**obj)
